/* #define DEBUG */
#include <stdio.h>
#define DIM 4
#include "../mas-slave.h"

/* Compile with mas-slave.h and linking with TOP-C library */

#define DIM 4
#define BANDS 2
#define BAND_SIZE ((DIM+BANDS-1)/BANDS) /* For b | n, this is just n/b */
int band_size = BAND_SIZE;
float matrix[DIM*DIM] = {20.0, 2.0, 3.0, 4.0, 5.0, 20.0, 7.0, 8.0,
                         9.0, 10.0, 20.0, 12.0, 13.0, 14.0, 15.0, 20.0};
/* Columns 0 up to first_unred_col[b] of band b are 0 */
int first_unred_col[BANDS];
int first_unred_band = 0; /* Done when first_unred_band = b */
int band_is_busy[BANDS];

#define max(x,y) ( (x) < (y) ? y : x )
struct band {
  int band_no;
  int first_unred_band;
  int first_unred_col;
  float band[BAND_SIZE * DIM];
};

void mytrace ( dest, msg )
  int dest;
  void *msg;
{ if (dest > 0) printf("band %d", *(int *)msg);
  else printf("row vector: %5.1f %5.1f ...",
	      (((struct band *)msg)->band)[0], ((struct band *)msg)->band[1] );
}
/* { if ( dest != 0 ) printf("band %d", *(int *)msg); } */

/* These will be set to DIM and BANDS inside main() */
int n, b;  /* n = matrix dimension; b = number of bands */

void reduce_row(matrix, row_to_red, row, pivot)
float *matrix; int row_to_red; int row; int pivot;
{ float scalar = matrix[row_to_red*n+pivot] / matrix[row*n+pivot];
  float *row1 = &(matrix[row_to_red*n]);
  float *row2 = &(matrix[row*n]);
  int j;
#ifdef DEBUG
  printf("reducing row %d, using row %d and pivot %d\n", row_to_red,row,pivot);
#endif
  row1[pivot] = 0;
  for (j = pivot+1; j < n; j++)
    row1[j] = row1[j] - scalar * row2[j];
}

void *get_task() {
  static int i;
  for (i = first_unred_band; i < b ; i++)
    if (first_unred_col[i] <= first_unred_col[max(first_unred_band-1,0)]
        && band_is_busy[i] == 0) {
      band_is_busy[i] = 1;
      return MSG(&i, sizeof(i)); }
  return NOTASK; }

void *do_task(band_ptr) /* 0 <= band_no <= b - 1 */
void *band_ptr;
{ int band_no = *(int *)band_ptr;
  int row_start = band_no * band_size;
  int row_end = (band_no + 1) * band_size;
  int row, i;
  int pivot = max(first_unred_col[max(first_unred_band-1,0)], 0);
  static struct band result;
  /* Will always be true:  get_task() satisfied this condition */
  if (first_unred_col[band_no] <= first_unred_col[max(first_unred_band-1,0)]) {
    for (row = max(first_unred_col[band_no], 0);
         row < first_unred_band * band_size; row++)
      for (i = row_start; i < row_end; i++)
        reduce_row(matrix, i, row, row);
    /* This is non-zero block at or below first unreduced,
       diagonal block; Diagonalize it in place */
    for (row = row_start; row < row_end; row++, pivot++)
      for (i = row + 1; i < row_end; i++)
        reduce_row(matrix, i, row, pivot);
    if (band_no == first_unred_band) first_unred_band++;
    first_unred_col[band_no] = first_unred_band * band_size; }
  result.band_no = band_no;
  result.first_unred_band = first_unred_band;
  result.first_unred_col = first_unred_col[band_no];
  for (i = 0; i < n * band_size; i++) /* copy band */
    result.band[i] = matrix[band_no*band_size*n + i];
  return MSG(&result, sizeof(result));
}

int get_task_result(res_ptr, band_ptr)
void *res_ptr; void *band_ptr;
{  return UPDATE;
}

void update_environment(res_ptr, band_ptr)
void *res_ptr; void *band_ptr;
{ struct band *result_ptr = res_ptr;
  float *mat_ptr = result_ptr->band;
  int i, band_no = *(int *)band_ptr;
  if (result_ptr->first_unred_band > first_unred_band)
    first_unred_band = result_ptr->first_unred_band;
  first_unred_col[result_ptr->band_no]
    = result_ptr->first_unred_col;
  for (i = band_no*band_size*n; i < (band_no+1)*band_size*n; i++)
    matrix[i] = *(mat_ptr++);
  band_is_busy[band_no] = 0; /* (Only master needs this) */
#ifdef DEBUG
if (is_master())
  { int i, j;
    printf("updating band %d\n", band_no);
    for (i = 0; i < DIM; i++, printf("\n"))
      for (j = 0; j < DIM; j++)
        printf("%5.1f ", matrix[i*DIM+j]);
  }
#endif
}

int main(argc, argv)
int argc;
char *argv[];
{ int i, j;
  n = DIM;  /* set dimension */
  b = BANDS;   /* number of bands */
  band_size = (n+b-1)/b; /* For b | n, this is just n/b */
  for (i = 0; i < b; i++) {
    /* Columns 0 until first_unred_col[i] of band i are 0 cols and
      next block is diagonalized; -1 means nothing diagonalized */
    first_unred_col[i] = -1;
    band_is_busy[i] = 0; }
/*
  matrix = malloc(n*n*sizeof(*matrix));
*/
  master_slave_trace = TRUE;
  master_slave_trace_cmd = mytrace;
  alarm(1200);
  init_master_slave(2, &argc, &argv);
  master_slave(get_task, do_task, get_task_result,
	       update_environment);
  if (is_master())
    for (i = 0; i < DIM; i++, printf("\n"))
      for (j = 0; j < DIM; j++)
        printf("%5.1f ", matrix[i*DIM+j]);
 exit(0);
 }

#if 0
/* THIS VERSION CAN BE USED TO DRIVE A SEQUENTIAL VERSION */
float matrix[DIM*DIM] = {20.0, 2.0, 3.0, 4.0, 5.0, 20.0, 7.0, 8.0,
                         9.0, 10.0, 20.0, 12.0, 13.0, 14.0, 15.0, 20.0};
int main()
{ int row, i, j;
  n = DIM; /* Used by reduce_row */
  void reduce_row(float *, int, int, int);
  for (row = 0; row < n; row++)
    for (i = row + 1; i < n; i++)
      reduce_row(matrix, i, row, row); /* row already reduced */
  for (i = 0; i < n; i++, printf("\n"))
    for (j = 0; j < n; j++)
      printf("%5.1f ", matrix[i*n+j]);
  exit(0);
}
#endif
