#include <topc.h>

/********************************************************************
 ********************************************************************
 ** ALGORITHM:  Euclidean sieve for factorization
 **   Demonstrates use of TOPC_raw_XXX
 **     for parallelizing legacy sequential code
 ** shared memory:  num_to_factor (divided by each factor found)
 ** task input:   input (test factor for num_to_factor)
 ** task output:  1 or 0 (true or false)
 ** update:  Divide num_to_factor by num
 ********************************************************************
 ********************************************************************/

#ifndef FALSE
enum {FALSE, TRUE};
#endif

// TOPC_trace = 0; /* Uncomment this, if no tracing is desired */
void trace_input( void *input ) { printf("%d", *(int *)input); } 
void trace_output( void *output ) { if (*(int *)output == 1) printf("TRUE");
                                    if (*(int *)output == 0) printf("FALSE");
                                  }
#define TASK_INTERVAL 1000

// int master_slave_trace;
// void (*master_slave_trace_cmd)(int dest, void *msg);


long factors[1000], max_factor=1, num_to_factor, task_res;
void MasSlaveFactor();

int main(int argc, char *argv[])
{
  int i;

  if (argc == 1 ) {
    printf("Missing number arg on command line.\n");
    exit(1); }
  TOPC_trace_input  = trace_input;
  TOPC_trace_output = trace_output;
  TOPC_init(&argc, &argv);

  MasSlaveFactor( atoi(argv[1]) );
  if ( TOPC_is_master() ) {
    for (i = 1; i <= factors[ 0 ]; i++) printf("%d ", factors[i]);
    printf("\n"); }
  TOPC_finalize();
  // printf("%d: Exiting\n", TOPC_rank()); fflush(stdout);
  exit(0);
}

TOPC_MSG_BUF DoTask( void *input )
{
  long limit;
  long num = *(long *)input;

  // printf("DoTask(%d)\n", num); fflush(stdout);
  limit = num + TASK_INTERVAL;
  for ( ; num < limit; num++ )
    if (num_to_factor % num == 0) {
      // printf("%d: DoTask returning 1\n", TOPC_rank()); fflush(stdout);
      task_res = 1;
      return TOPC_MSG(&task_res, sizeof(long));
    }
  // printf("%d: DoTask returning 0\n", TOPC_rank()); fflush(stdout);
  task_res = 0;
  return TOPC_MSG(&task_res, sizeof(long));
}

TOPC_ACTION CheckTaskResult( void *input, void *output )
{
  // printf("CheckTaskResult()\n"); fflush(stdout);
  if ( *(long*)output == 0 ) return NO_ACTION;
  // NOTE: Suppose we factor 12, 2 is outstanding, and 4 is reported to factor
  if ( *(long*)input < max_factor ) return UPDATE;
  if ( ! TOPC_is_up_to_date() ) return REDO;
  return UPDATE;
}

void UpdateSharedMemory( void *input, void *output )
{
  long limit, num = *(long *)input;
  int i;
  // printf("%d: UpdateSharedMemory()\n", TOPC_rank()); fflush(stdout);
  limit = num + TASK_INTERVAL;
  // NOTE: Suppose we factor 12, 4 is reported as factor before 2 is reported
  for ( ; num < limit; num++ ) {
    if ( num_to_factor % num == 0 ) {
      if ( num < max_factor ) {
        max_factor = 1; // re-compute max_factor
        for( i = 1; i <= factors[0]; i++ ) {
          if (factors[i] > max_factor) max_factor = factors[i];
          while (factors[i] % num == 0) {
            factors[i] = factors[i] / num;
            num_to_factor = num_to_factor * num;
          }
        }
      }
      while (num_to_factor % num == 0) {
        factors[ ++factors[0] ] = num;
        num_to_factor = num_to_factor / num;
        if (num_to_factor == 1) return;
      }
    }
  }
}

void MasSlaveFactor( long num )
{ TOPC_MSG_BUF input;
  long last_num;

  factors[0]=0;
  num_to_factor = num;
  //printf("Entering master_slave()\n"); fflush(stdout);
  TOPC_begin_raw_master_slave(DoTask, CheckTaskResult,
                              UpdateSharedMemory);
  for ( last_num = 2; last_num <= num_to_factor; last_num += TASK_INTERVAL )
    TOPC_raw_submit_task_input(TOPC_MSG(&last_num, sizeof(last_num)));
  TOPC_end_raw_master_slave();
}
